package jp.ne.motoki.android.hellolivewallpaper;

import static jp.ne.motoki.android.hellolivewallpaper.Utils.log_warn;

import java.io.FileNotFoundException;
import java.io.InputStream;
import java.nio.Buffer;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.FloatBuffer;

import javax.microedition.khronos.opengles.GL10;
import javax.microedition.khronos.opengles.GL11;

import android.content.Context;
import android.content.SharedPreferences;
import android.graphics.Bitmap;
import android.graphics.BitmapFactory;
import android.net.Uri;
import android.opengl.GLU;
import android.opengl.GLUtils;
import android.preference.PreferenceManager;

public class GLEngineRender {

    /**
     * float 型のbyte数。
     */
    private static final int SIZE_OF_FLOAT = Float.SIZE / 8;
    
    /**
     * 頂点座標のbyte数。
     */
    private static final int SIZE_OF_VERTEX_POINTER = 3;
    
    /**
     * UV座標のbyte数。
     */
    private static final int SIZE_OF_TEX_COORD_POINTER = 2;
    
    /**
     * １つの頂点情報のbyte数。
     */
    private static final int STRIDE =
            SIZE_OF_FLOAT * (SIZE_OF_VERTEX_POINTER + SIZE_OF_TEX_COORD_POINTER);
    
    /**
     * 頂点座標のオフセット。
     */
    private static final int OFFSET_OF_VERTEX_POINTER = 0;
    
    /**
     * UV座標のオフセット。
     */
    private static final int OFFSET_OF_TEX_COORD_POINTER =
            SIZE_OF_FLOAT * SIZE_OF_VERTEX_POINTER;
    
    private final float[] VERTICES = {
            // x      y     z     u     v
            -0.5f,  0.5f, 0.0f, 0.0f, 0.0f, // !< 左上
            -0.5f, -0.5f, 0.0f, 0.0f, 1.0f, // !< 左下
             0.5f,  0.5f, 0.0f, 1.0f, 0.0f, // !< 右上
             0.5f, -0.5f, 0.0f, 1.0f, 1.0f, // !< 右下
    };
    
    private final Context context;

    private int windowWidth = -1, windowHeight = -1;

    private int vertexBufferObject = 0;

    private int textureID = 0;
    
    /**
     * テクスチャの幅・高さ
     */
    private int textureWidth = -1, textureHeight = -1;

    private float rotateY = 0.0f;
    private float rotateX = 0.0f;
    
    public GLEngineRender(Context context) {
        if (context == null) {
            throw new NullPointerException("context == null");
        }
        this.context = context;
    }

    public void onSurfaceChanged(
            GL10 gl10, int windowWidth, int windowHeight) {
        this.windowWidth = windowWidth;
        this.windowHeight = windowHeight;

        // ! 事前に転送済みにしておく
        transferVBO(gl10);
        
        //! texture
        bindTexture(gl10);
    }

    public void onDrawFrame(GL10 gl10) {
        gl10.glClearColor(0, 1, 1, 1);
        gl10.glClear(GL10.GL_COLOR_BUFFER_BIT);
        
        setTextureArea(gl10, 0, 0, textureWidth, textureHeight);
        
        float widthScale = (float) windowWidth / (float) textureWidth;
        float heightScale = (float) windowHeight / (float) textureHeight;
        float scale = Math.min(widthScale, heightScale);
        int width = (int) (textureWidth * scale);
        int height = (int) (textureHeight * scale);
        drawQuad(gl10, (windowWidth - width) / 2, (windowHeight - height) / 2, width, height);
    }
    
    private void transferVBO(GL10 gl10) {
        ByteBuffer bb = ByteBuffer.allocateDirect(VERTICES.length * SIZE_OF_FLOAT);
        bb.order(ByteOrder.nativeOrder());
        FloatBuffer vbo = bb.asFloatBuffer();
        vbo.put(VERTICES);
        vbo.position(0);

        gl10.glEnableClientState(GL10.GL_TEXTURE_COORD_ARRAY);
        gl10.glEnableClientState(GL10.GL_VERTEX_ARRAY);

        createVBO(gl10, vbo);
    }
    
    private void createVBO(GL10 gl10, Buffer vbo) {
        GL11 gl11 = (GL11) gl10;
        int[] buffers = new int[1];
        gl11.glGenBuffers(1, buffers, 0);
        vertexBufferObject = buffers[0];
        gl11.glBindBuffer(GL11.GL_ARRAY_BUFFER, vertexBufferObject);
        gl11.glBufferData(
                GL11.GL_ARRAY_BUFFER, vbo.capacity( ) * SIZE_OF_FLOAT,
                vbo, GL11.GL_STATIC_DRAW );
        gl11.glVertexPointer(
                SIZE_OF_VERTEX_POINTER, GL10.GL_FLOAT, STRIDE,
                OFFSET_OF_VERTEX_POINTER);
        gl11.glTexCoordPointer(
                SIZE_OF_TEX_COORD_POINTER, GL10.GL_FLOAT, STRIDE,
                OFFSET_OF_TEX_COORD_POINTER);
        gl11.glBindBuffer( GL11.GL_ARRAY_BUFFER, 0 );
    }
    
    private void bindTexture(GL10 gl10) {
        String uriString = getImageUriString();
        if (uriString == null) {
            // TODO 仮実装
            log_warn("uriString == null");
            return;
        }
        
        Uri imageUri = Uri.parse(uriString);
        Bitmap bitmap = null;
        try {
            bitmap = createBitmap(imageUri);
        } catch (FileNotFoundException e) {
            // TODO 仮実装
            log_warn("createBitmap failed");
            return;
        }

        gl10.glEnable(GL10.GL_TEXTURE_2D);
        int[] buffer = new int[1];
        gl10.glGenTextures(1, buffer, 0);
        textureID = buffer[0];
        textureWidth = bitmap.getWidth();
        textureHeight = bitmap.getHeight();

        gl10.glBindTexture(GL10.GL_TEXTURE_2D, textureID);
        GLUtils.texImage2D(GL10.GL_TEXTURE_2D, 0, bitmap, 0);
        gl10.glTexParameterf(
                GL10.GL_TEXTURE_2D, GL10.GL_TEXTURE_MIN_FILTER,
                GL10.GL_NEAREST);
        gl10.glTexParameterf(
                GL10.GL_TEXTURE_2D, GL10.GL_TEXTURE_MAG_FILTER,
                GL10.GL_NEAREST);

        bitmap.recycle( );
    }

    private void drawQuad(GL10 gl10, int x, int y, int w, int h) {
        float sizeX = ( float ) w / ( float ) windowWidth * 2;
        float sizeY = ( float ) h / ( float ) windowHeight * 2;
        float sx = ( float ) x / ( float ) windowWidth * 2;
        float sy = ( float ) y / ( float ) windowHeight * 2;
        gl10.glMatrixMode(GL10.GL_MODELVIEW);
        gl10.glLoadIdentity();
        gl10.glTranslatef(
                -1.0f + sizeX / 2.0f + sx, 1.0f - sizeY / 2.0f - sy, 0.0f);
        {
            gl10.glRotatef(rotateY++, 0, 1, 0);
            gl10.glRotatef(rotateX, 1, 0, 0);
            rotateX += 2.0f;
        }
        gl10.glScalef(sizeX, sizeY, 1.0f);
        
        // カメラ転送
        transfarToCamera(gl10);
        
        gl10.glDrawArrays(GL10.GL_TRIANGLE_STRIP, 0, 4);
    }
    
    // TODO 仮実装
    private void transfarToCamera(GL10 gl10) {
        gl10.glMatrixMode(GL10.GL_PROJECTION);
        gl10.glLoadIdentity();
        GLU.gluPerspective(gl10, 45.0f, /*aspect*/1  , 0.01f, 100.0f);
        GLU.gluLookAt(gl10, 0, 0, 10.0f, 0, 0, 0.0f, 0.0f, 1.0f, 0.0f);
    }

    private void setTextureArea(GL10 gl10, int x, int y, int w, int h) {
        float tw = ( float ) w / ( float ) textureWidth;
        float th = ( float ) h / ( float ) textureHeight;
        float tx = ( float ) x / ( float ) textureWidth;
        float ty = ( float ) y / ( float ) textureHeight;

        gl10.glMatrixMode(GL10.GL_TEXTURE);
        gl10.glLoadIdentity();
        gl10.glTranslatef(tx, ty, 0.0f);
        gl10.glScalef(tw, th, 1.0f);
    }

    private String getImageUriString() {
        SharedPreferences sharedPreferences =
            PreferenceManager.getDefaultSharedPreferences(context);
        return sharedPreferences.getString(
                context.getString(R.string.key_image), null);
    }
    
    private Bitmap createBitmap(Uri imageUri) throws FileNotFoundException {
        InputStream is = 
                context.getContentResolver().openInputStream(imageUri);
        return BitmapFactory.decodeStream(is);
    }
    
}